import codecs
import urllib.parse

from ..classes.StiParameter import StiParameter
from ..enums.StiDatabaseType import StiDatabaseType
from ..enums.StiDataCommand import StiDataCommand
from ..enums.StiFrameworkType import StiFrameworkType
from ..events.StiDataEventArgs import StiDataEventArgs
from ..events.StiEvent import StiEvent
from ..StiDataAdapter import StiDataAdapter
from .StiBaseRequest import StiBaseRequest
from .StiBaseResponse import StiBaseResponse
from .StiBaseResult import StiBaseResult


class StiBaseHandler:
    """
    Event handler for all requests from the report generator. 
    The incoming request is processed, a data adapter is created and all necessary actions are performed.
    """

    version: str = '2024.2.1'
    checkDataAdaptersVersion: bool = True
    framework: str = StiFrameworkType.DEFAULT
    origin: str = None
    query: dict = None
    body: str = None
    error: str = None
    request: StiBaseRequest = None
    dataAdapter: StiDataAdapter = None


### Private

    __url: str = None


### Properties
    
    @property
    def url(self):
        return '' if self.__url == None else self.__url

### Events

    onBeginProcessData: StiEvent = None
    """The event is invoked before data request, which needed to render a report."""

    onEndProcessData: StiEvent = None
    """The event is invoked after loading data before rendering a report."""


### Private

    def __setQuery(self, query: dict | str) -> None:
        if query != None:
            self.query = query if isinstance(query, dict) else dict(urllib.parse.parse_qsl(query))

    def __setBody(self, body: bytes | str) -> None:
        if body != None:
            self.body = body if type(body) == str else codecs.decode(body)

    def __processFrameworkRequest(self, request: object) -> bool:
        try:
            from django.http import HttpRequest as DjangoRequest
            if isinstance(request, DjangoRequest):
                self.__setQuery(request.GET)
                self.__setBody(request.body)
                self.origin = '{0}://{1}'.format(request.scheme, request.get_host())
                self.framework = StiFrameworkType.DJANGO
                return True
        except Exception as e:
            self.error = 'Request: ' + str(e)
            return False

        try:
            from flask import Request as FlaskRequest
            if isinstance(request, FlaskRequest):
                self.__setQuery(request.args.to_dict())
                self.__setBody(request.get_data(False))
                self.origin = request.origin
                self.framework = StiFrameworkType.FLASK
                return True
        except Exception as e:
            self.error = 'Request: ' + str(e)
            return False

        try:
            from tornado.httputil import HTTPServerRequest as TornadoRequest
            if isinstance(request, TornadoRequest):
                self.__setQuery(request.query)
                self.__setBody(request.body)
                self.origin = '{0}://{1}'.format(request.protocol, request.host)
                self.framework = StiFrameworkType.TORNADO
                return True
        except Exception as e:
            self.error = 'Request: ' + str(e)
            return False
        
        return True

    def __getSupportedDataAdaptersResult(self) -> StiBaseResult:
        result = StiBaseResult.getSuccess()
        result.types = StiDatabaseType.getTypes()
        result.handlerVersion = self.version
        return result
    
    def __processParameters(self):
        parameters = dict()
        if len(self.request.queryString or '') > 0 and len(self.request.parameters or '') > 0:
            for item in self.request.parameters:
                name = item['name'][1:] if item['name'][0] == '@' or item['name'][0] == ':' else item['name']
                parameters[name] = StiParameter(name, item['type'], item['typeName'], item['typeGroup'], item['size'], item['value'])

        self.request.parameters = parameters
    
    def __getDataAdapterResult(self):
        args = StiDataEventArgs(self.request)
        self.onBeginProcessData(args)

        self.dataAdapter = StiDataAdapter.getDataAdapter(args.database, args.connectionString)
        if self.dataAdapter == None:
            return StiBaseResult.getError(f'Unknown database type: {args.database}')

        if self.request.command == StiDataCommand.RETRIEVE_SCHEMA:
            args.result = self.dataAdapter.executeQuery(args.dataSource)
            self.onEndProcessData(args)
            return args.result

        if self.request.command == StiDataCommand.EXECUTE or self.request.command == StiDataCommand.EXECUTE_QUERY:
            try:
                from ..StiMongoDbAdapter import StiMongoDbAdapter
                if (isinstance(self.dataAdapter, StiMongoDbAdapter)):
                    args.queryString = args.dataSource
            except:
                pass

            if self.request.command == StiDataCommand.EXECUTE:
                args.queryString = self.dataAdapter.makeQuery(args.queryString, args.parameters)

            if len(args.parameters) > 0:
                args.queryString = StiDataAdapter.applyQueryParameters(args.queryString, args.parameters, self.request.escapeQueryParameters)

            args.result = self.dataAdapter.executeQuery(args.queryString)
            self.onEndProcessData(args)
            return args.result
        
        return self.dataAdapter.test()
    

### Protected

    def _createRequest(self):
        return StiBaseRequest()
    
    def _checkEvent(self):
        return self.request.event == None or self.request.event == 'BeginProcessData'
    
    def _checkCommand(self):
        commands = [getattr(StiDataCommand, field) for field in dir(StiDataCommand) if not callable(getattr(StiDataCommand, field)) and not field.startswith('_')]
        if self.request.command in commands:
            return True
        
        return False

### Public

    def processRequest(self, request: object = None, query: dict | str = None, body: bytes | str = None) -> bool:
        """The request process, reading all passed parameters, retrieving data from the database, and other necessary actions of the event handler."""

        self.error = None

        if request != None:
            self.__processFrameworkRequest(request)
            if (self.error != None):
                return False
        
        if self.framework == StiFrameworkType.DEFAULT:
            try:
                self.__setQuery(query)
                self.__setBody(body)
            except Exception as e:
                self.error = 'Request: ' + str(e)
                return False
        
        self.request = self._createRequest()
        self.request.process(self.query, self.body)
        if len(self.request.error or '') > 0:
            self.error = self.request.error
            return False
        
        if not self._checkEvent():
            self.error = f'Unknown event: {self.request.command}'
            return False
        
        if not self._checkCommand():
            self.error = f'Unknown command: {self.request.command}'
            return False
        
        return True

    def getFrameworkResponse(self, handler = None) -> object:
        """
        Returns a response intended for one of the supported frameworks.
        """

        return self.getResponse().getFrameworkResponse(handler)

    def getResponse(self) -> StiBaseResponse:
        """
        The result of executing an event handler request.
        You can get the data, its type and other parameters necessary to create a web server response.
        """
        
        return StiBaseResponse(self)

    def getResult(self):
        """
        The result of executing an event handler request. 
        The result contains a collection of data, message about the result of the command execution, and other technical information.
        """

        if len(self.error or '') > 0:
            result = StiBaseResult.getError(self.error)
            result.handlerVersion = self.version
            return result
        
        if self.request.command == StiDataCommand.GET_SUPPORTED_ADAPTERS:
            return self.__getSupportedDataAdaptersResult()

        self.__processParameters()
        result = self.__getDataAdapterResult()
        result.handlerVersion = self.version
        return result
    

### Constructor

    def __init__(self, url: str = None):
        self.onBeginProcessData = StiEvent(self, 'onBeginProcessData')
        self.onEndProcessData = StiEvent(self, 'onEndProcessData')
        self.__url = url


class StiConnectionInfo:
    """All database connection information taken from the connection string."""

    driver: str = ''
    host: str = ''
    port: int = 0
    database: str = ''
    userId: str = ''
    password: str = ''
    charset: str = ''
    privilege: str = ''
    dataPath: str = ''
    schemaPath: str = ''


from ..classes.StiDataResult import StiDataResult
from ..classes.StiParameter import StiParameter
from .StiEventArgs import StiEventArgs


class StiDataEventArgs(StiEventArgs):

    command: str = None
    """The current command for the data adapter. The supported commands are in the 'StiDataCommand' enum."""

    database: str = None
    """The database type for which the command will be executed. The supported database types are in the 'StiDatabaseType' enum."""

    connection: str = None
    """The name of the current database connection."""

    dataSource: str = None
    """The name of the current data source."""

    connectionString: str = None
    """The connection string for the current data source."""

    queryString: str = None
    """The SQL query that will be executed to get the data array of the current data source."""

    parameters: dict[str, StiParameter] = None
    """A set of parameters for the current SQL query."""

    result: StiDataResult = None
    """The result of executing an event handler request."""


from typing import Final


class StiDatabaseType:
    MYSQL: Final = 'MySQL'
    MSSQL: Final = 'MS SQL'
    POSTGRESQL: Final = 'PostgreSQL'
    FIREBIRD: Final = 'Firebird'
    ORACLE: Final = 'Oracle'
    ODBC: Final = 'ODBC'
    MONGODB: Final = 'MongoDB'

    def getTypes():
        return [getattr(StiDatabaseType, field) for field in dir(StiDatabaseType) if not callable(getattr(StiDatabaseType, field)) and not field.startswith('__')]


from __future__ import annotations

import codecs
import json
import typing

from ..enums.StiFrameworkType import StiFrameworkType
from .StiBaseResult import StiBaseResult

if typing.TYPE_CHECKING:
    from .StiBaseHandler import StiBaseHandler


class StiBaseResponse:
    """
    The result of executing an event handler request. 
    You can get the data, its type and other parameters necessary to create a web server response.
    """

    handler: StiBaseHandler = None
    result: StiBaseResult = None


### Private

    def __getArrtibutes(obj: object):
        return { attr: getattr(obj, attr) for attr in dir(obj) if not callable(getattr(obj, attr)) and not attr.startswith('_') }


### Properties

    @property
    def origin(self) -> str:
        """Returns the detected origin url for the handler response. Can be used for the 'Access-Control-Allow-Origin' header of the response."""

        return self.handler.origin

    @property
    def mimetype(self) -> str:
        """Returns the mime-type for the handler response."""

        return 'application/json'
    
    @property
    def contentType(self) -> str:
        """Returns the content type for the handler response. Can be used for the 'Content-Type' header of the response."""

        return self.mimetype + '; charset=utf-8'
    
    @property
    def data(self) -> bytes:
        """Returns the handler response as a byte array. When using encryption, the response will be encrypted and encoded into a Base64 string."""

        data = json.dumps(self.result, default = StiBaseResponse.__getArrtibutes)
        if self.handler.request.encryptData == True:
            data = codecs.encode(codecs.encode(data.encode(), 'base64').decode(), 'rot13')
        return data.encode()

### Public

    def getFrameworkResponse(self, handler = None):
        """
        Returns a response intended for one of the supported frameworks.
        """
        
        if self.handler.framework == StiFrameworkType.DJANGO:
            try:
                from django.http import HttpResponse
                contentType = self.contentType
                response = HttpResponse(self.data, content_type = contentType)
                return response
            except:
                return self

        if self.handler.framework == StiFrameworkType.FLASK:
            try:
                from flask import make_response
                response = make_response(self.data)
                response.mimetype = self.mimetype
                response.headers.add('Access-Control-Allow-Origin', self.origin)
                return response
            except:
                return self
            
        if self.handler.framework == StiFrameworkType.TORNADO:
            try:
                from tornado.web import RequestHandler as TornadoRequestHandler
                if isinstance(handler, TornadoRequestHandler):
                    handler.set_header('Content-Type', self.contentType)
                    handler.write(self.data)
                    return None

                return self
            except:
                return self

        return self
    

### Constructor

    def __init__(self, handler: StiBaseHandler, result: StiBaseResult = None):
        self.handler = handler
        self.result = result if result != None else self.handler.getResult()



from .classes.StiDataResult import StiDataResult
from .StiDataAdapter import StiDataAdapter


class StiFirebirdAdapter(StiDataAdapter):
    version: str = '2024.2.1'
    checkVersion: bool = True

    def getOdbcConnectionString(self):
        connectionString: str = \
            f'Driver={self.connectionInfo.driver};' \
            f'Dbname={self.connectionInfo.host}/{self.connectionInfo.port}:{self.connectionInfo.database};' \
            f'Uid={self.connectionInfo.userId};' \
            f'Pwd={self.connectionInfo.password};'
        
        return connectionString

    def connect(self):
        if self.connectionInfo.driver:
            return self.connectOdbc()
        
        if not self.connectionInfo.charset:
            self.connectionInfo.charset = 'utf8'
        
        try:
            from firebird.driver import connect
            dsn = f'{self.connectionInfo.host}/{self.connectionInfo.port}:{self.connectionInfo.database}'
            self.connectionLink = connect(
                user = self.connectionInfo.userId,
                password = self.connectionInfo.password,
                database = dsn,
                charset = self.connectionInfo.charset)
        except Exception as e:
            return StiDataResult.getError(self, str(e))
        
        return StiDataResult.getSuccess(self)
    
    def process(self):
        if not super().process():
            return False

        self.connectionInfo.port = 3050

        parameterNames = {
            'driver': ['driver'],
            'host': ['server', 'host', 'location', 'datasource', 'data source'],
            'port': ['port'],
            'database': ['database', 'dbname'],
            'userId': ['uid', 'user', 'username', 'userid', 'user id'],
            'password': ['pwd', 'password'],
            'charset': ['charset']
        }

        return self.parseParameters(parameterNames)
    
    def makeQuery(self, procedure: str, parameters: list):
        paramsString = super().makeQuery(procedure, parameters)
        return f'EXECUTE PROCEDURE {procedure} {paramsString}'
    


class StiEventArgs:

    event: str = None
    """Name of the current event."""
    
    sender: object = None
    """The component that sent the request."""


### Private

    def __populateVars(self, obj: object):
        attributes = [attr for attr in dir(self) if not attr.startswith('_') and attr != 'sender']
        for attr in attributes:
            if hasattr(obj, attr):
                setattr(self, attr, getattr(obj, attr))


### Constructor

    def __init__(self, obj: object = None):
        if (obj != None):
            self.__populateVars(obj)


import pyodbc
from pyodbc import Connection

from .classes.StiDataResult import StiDataResult
from .StiDataAdapter import StiDataAdapter


class StiOdbcAdapter(StiDataAdapter):
    version: str = '2024.2.1'
    checkVersion: bool = True
    connectionLink: Connection

    def connect(self):
        try:
            self.connectionLink = pyodbc.connect(self.connectionString)
            if self.connectionInfo.charset:
                self.connectionLink.setdecoding(pyodbc.SQL_CHAR, self.connectionInfo.charset)
                self.connectionLink.setdecoding(pyodbc.SQL_WCHAR, self.connectionInfo.charset)
        except Exception as e:
            return StiDataResult.getError(self, str(e))
        
        return StiDataResult.getSuccess(self)
    
    def process(self):
        if not super().process():
            return False

        parameterNames = {
            'driver': ['driver'],
            'userId': ['uid', 'user', 'username', 'userid', 'user id'],
            'password': ['pwd', 'password']
        }

        return self.parseParameters(parameterNames)
    
    


import codecs
import json

from .StiBaseResult import StiBaseResult


class StiBaseRequest:
    """Contains all set request parameters passed to the event handler."""

    event: str = None
    command: str = None
    encryptData: bool = False
    connectionString: str = None
    queryString: str = None
    parameters: dict[str, object] = None
    database: str = None
    dataSource: str = None
    connection: str = None
    timeout: int = 0
    escapeQueryParameters: bool = False
    error: str = None


### Private

    def __populateVars(self, obj: object, prefix: bool = False):
        fields = dir(self)
        for prop in obj:
            field = prop[4:] if prefix and prop[0:4] == 'sti_' else prop
            if field in fields:
                self._setField(field, obj[prop])


### Protected

    def _setField(self, name, value):
        setattr(self, name, value)


### Public

    def process(self, query: dict, body: str) -> bool:
        if len(query) > 0:
            self.__populateVars(query, True)

        if len(body) > 0:
            if body[0] != '{':
                try:
                    body = codecs.decode(body, 'rot_13')
                    body = codecs.decode(body.encode(), 'base64').decode()
                except Exception as e:
                    self.error = 'Base64: ' + str(e)
                    return False
                
                self.encryptData = True

            try:
                obj = json.loads(body)
            except Exception as e:
                self.error = 'JSON: ' + str(e)
                return False

            self.__populateVars(obj)

        return True

    def getResult(self):
        """
        The result of executing an event handler request. 
        The result contains a collection of data, message about the result of the command execution, and other technical information.
        """

        if self.error != None:
            return StiBaseResult.getError(self.error)

        StiBaseResult.getSuccess()
    


from typing import Final


class StiDataCommand:
    GET_SUPPORTED_ADAPTERS: Final = 'GetSupportedAdapters'
    TEST_CONNECTION: Final = 'TestConnection'
    RETRIEVE_SCHEMA: Final = 'RetrieveSchema'
    EXECUTE: Final = 'Execute'
    EXECUTE_QUERY: Final = 'ExecuteQuery'
